#! /usr/bin/env perl

#Copyright 2008 Cold Spring Harbor Laboratory.
# Written by Yaniv Erlich (erlich@cshl.edu)



package Cluster::Ipar_formatter;
require Exporter;
our @EXPORT = qw(quick_formatter);


use strict;
use warnings;

use FindBin qw($Bin);

use Carp;
use Readonly;
use Data::Dumper;
use File::Basename;
use File::Copy;
use Cwd qw(abs_path);
use Getopt::Long;
use Data::Dumper;
use Pod::Usage;
use File::Path;



use SGE_API;
use Submit_tools;
use Snapshot;






Readonly my $D_GZ_SUFFIX                  => ".gz";
Readonly my $D_GUNZIP                     => 'gunzip';
Readonly my $D_GZIP                       => 'gzip';
Readonly my $SUFFIX_POS 		  => 'pos.txt';
Readonly my $SUFFIX_IPAR_INT              => 'int.txt.p';
Readonly my $D_IPAR_PREFIX                => 'ipar';
Readonly my $D_POS_PREFIX                 => 'pos';


Readonly my $D_SUFFIX_OF_INT_FILES        => "$Snapshot::D_SUFFIX_OF_INT_FILES";
Readonly my $D_SYMBOL                     => "$Snapshot::D_SYMBOL";
Readonly my $D_LIFETIME                             => '10h';
Readonly my $D_REFRESH_TIME                         => '0.5m';
Readonly my $D_SOFT_LINK_FOLDER                     => 'Soft_links/';
Readonly my $DOT                                    => q{.};
Readonly my $WILD_CARD_SYMBOL                       => '<X>';
Readonly my $D_OUPUT_DIR                => './';
Readonly my $LOCATION_OF_MAN_SCRIPTS                => "$Bin/Cluster/Remote_Scripts"; #Cluster
Readonly my $D_REMOTE_SCRIPT_NAME                   => "$LOCATION_OF_MAN_SCRIPTS/"."ipar_to_int.pl";

#^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^




#######################################################################################################################


sub quick_formatter {

    my ($r_arg) = @_;


    my $path_to_ipar_files = $r_arg->{'path_to_ipar_files'};
    my $path_to_pos_files  = $r_arg->{'path_to_pos_files'};
    my $r_tiles            = $r_arg->{'r_tiles'};
    my $r_lanes            = $r_arg->{'r_lanes'};
    my $zipped             = $r_arg->{'zipped'};
    my $verbose            = $r_arg->{'verbose'};
    my $debug              = $r_arg->{'debug'};
    my $output_dir         = $r_arg->{'output_dir'};


    $output_dir = abs_path $output_dir;

    foreach my $lane (@$r_lanes) {

        my ($r_i,$r_p) = 
            get_filenames_of_int_and_pos (
                                          $path_to_ipar_files,
                                          $path_to_pos_files,
                                          $lane,
                                          $r_tiles,
                                          $zipped
                                         );

        my ($r_soft_files,
            $ipar_template,
            $pos_template,
            $path_to_soft_links) =    create_softlink_to_files($r_i,$r_p,$lane,$zipped,$output_dir);


        send_to_cluster($r_soft_files,$ipar_template,$pos_template,$zipped,$verbose,$debug,$lane,$output_dir,$r_tiles);

        clean_soft_links($path_to_soft_links, $debug);
    }
}

#################################################################################################################
sub send_to_cluster {

    my ($r_soft_files,$ipar_template,$pos_template,$zipped,$verbose,$debug,$lane,$output_dir,$r_tiles) = @_;

    $output_dir = abs_path $output_dir;

    my @tiles = @$r_tiles;

    my @starts =  $tiles[0];
    my @gaps   =  1;
    my $last_tile = $tiles[scalar @tiles-1];
    my @stops  =  $last_tile++;



    print "Formatting IPAR files of lane:$lane\n",
        if ($verbose);



    my $cluster_script = SGE_API->new();
    my ($counter,undef) = $cluster_script->set_range(\@starts,\@stops,\@gaps);

    print "Total number of permutations is $counter\n" if ($debug);



    $cluster_script->add_required_file({refer=> 'ipar_to_int',
                                        name => $D_REMOTE_SCRIPT_NAME,
                                        place=> 'local'});


     $cluster_script->add_required_file({refer=> 'ipar',
                                         name => $ipar_template,
                                         place=> 'local'});


     $cluster_script->add_required_file({refer=> 'pos_file',
                                         name => $pos_template,
                                         place=> 'local'});


    my $command_line = join (' ',
                                 '<ipar_to_int>',
                                 '--ipar_file',                  '<ipar>',
                                 '--pos_file',                    '<pos_file>',
                                 '--lane',                        $lane,
                                 '--output_dir'            ,      $output_dir,
                                 '--tile_number'           ,      $WILD_CARD_SYMBOL,
                            );


    if ($zipped) {
        $command_line .= " --zipped"
    }


    print "$command_line\n" if ($verbose);

    $cluster_script->add_command($command_line);
    my $script_to_submit = $cluster_script->compile;
    $cluster_script->check_integrity;


     print "Generating script to send to the cluster: $script_to_submit\n" if ($debug);
     if ($debug) {_print_script($script_to_submit)};

     my $qsub = Submit_tools->new();
     $qsub->life_time($D_LIFETIME);
     $qsub->range(1,scalar @{$r_soft_files->[0]},1);
     $qsub->submit($script_to_submit);
     my $job_name = $qsub->get_job_id;
     my @stdout_files = $qsub->get_message_filenames;
     $qsub->automatic_pilot($D_REFRESH_TIME);
     $cluster_script->bye;

}








sub create_softlink_to_files {

    my ($r_ipar_files_to_read,$r_pos_files_to_read,$lane,$zipped,$output_dir) = @_;


    confess "error in create_softlink_to_files"
        if (scalar @$r_ipar_files_to_read !=
            scalar @$r_pos_files_to_read);

    my @soft_links;
    if (not -d "$output_dir/$D_SOFT_LINK_FOLDER") {
        mkdir "$output_dir/$D_SOFT_LINK_FOLDER" || confess "can't create $output_dir/$D_SOFT_LINK_FOLDER: $!";
    }
    my $path = abs_path "$output_dir/$D_SOFT_LINK_FOLDER";





    #GO_OVER_FILES:
    my $i = 0;
    foreach  (0..(scalar @$r_ipar_files_to_read)-1) {

        $i++;

        my $ipar_file       = $r_ipar_files_to_read->[$_];
        my $pos_file        = $r_pos_files_to_read->[$_];


        my $soft_link_ipar = join $DOT, $D_IPAR_PREFIX, $lane, $i;
        my $soft_link_pos  = join $DOT, $D_POS_PREFIX, $lane, $i;


        if ($zipped) {
            $soft_link_ipar .= $D_GZ_SUFFIX;
        }


        $soft_link_ipar = join '/', $path, $soft_link_ipar;
        $soft_link_pos  = join '/', $path, $soft_link_pos;

        $soft_link_ipar = abs_path $soft_link_ipar;
        $soft_link_pos  = abs_path $soft_link_pos;


        if ((not -e $ipar_file) or (not -e $pos_file)) {
            confess "ipar file:$ipar_file or pos file: $pos_file does not exist\n";
        }

        symlink $ipar_file,$soft_link_ipar 
            || confess "can't create soft line $soft_link_ipar=>$ipar_file: $!\n";

        symlink $pos_file,$soft_link_pos 
            || confess "can't create soft line $pos_file=>$soft_link_pos: $!\n";


        push @{$soft_links[0]}, $soft_link_ipar;
        push @{$soft_links[1]}, $soft_link_pos;

    }

    my $ipar_template = join $DOT, $D_IPAR_PREFIX, $lane, $WILD_CARD_SYMBOL;
    my $pos_template  = join $DOT, $D_POS_PREFIX , $lane, $WILD_CARD_SYMBOL;

    if ($zipped) {
        $ipar_template .= $D_GZ_SUFFIX;
    }

    $ipar_template = join '/', $path, $ipar_template;
    $pos_template  = join '/', $path, $pos_template;

    return (\@soft_links,$ipar_template,$pos_template,$path);


}




sub clean_soft_links {
    my ($path_to_soft_links, $debug) = @_;

    rmtree($path_to_soft_links, {verbose => $debug});
    return 1;
}











sub get_filenames_of_int_and_pos {

    my (
        $path_to_ipar_files,
        $path_to_pos_files,
        $lane,
        $r_tiles,
        $zipped,
       ) = @_;



    my @ipar_files_fullname = 
        Snapshot::get_filename_in_illumina_format({ path   => $path_to_ipar_files,
                                                    lane    => $lane,
                                                    tiles   => $r_tiles,
                                                    suffix  => $SUFFIX_IPAR_INT,
                                                    symbol  => $D_SYMBOL});

    my @pos_files_fullname =
        Snapshot::get_filename_in_illumina_format({ path   => $path_to_pos_files,
                                                    lane    => $lane,
                                                    tiles   => $r_tiles,
                                                    suffix  => $SUFFIX_POS,
                                                    symbol  => $D_SYMBOL});


    if ($zipped) {
       @ipar_files_fullname = map {$_."$D_GZ_SUFFIX"} @ipar_files_fullname;
    }

    return (\@ipar_files_fullname,\@pos_files_fullname);

}


sub _convert_gz_to_text_file {
    my ($gz_file) = @_;
    my $gz_basename = basename $gz_file;
    copy($gz_file,"/tmp/$gz_basename") or confess "can't copy $gz_file: $!\n";
    $gz_file = "/tmp/$gz_basename";
    system "$D_GUNZIP $gz_file";
    sleep (5);
    $gz_file =~ s/$D_GZ_SUFFIX//;
    return $gz_file;
}




sub _print_script {

    my $filename = shift;
    print "\n",
          "-------------------------------------------\n",
    open my $FH,'<', $filename
        or confess "can't open $filename: $!\n";

    while (<$FH>) {
        print "$_";
    }
    print "-------------------------------------------\n",
          "\n";
    close $FH;
    return 1;
}

1;
