#Copyright 2008 Cold Spring Harbor Laboratory.
# Written by Yaniv Erlich (erlich@cshl.edu)

#This package samples Solexa data (intensity values and sequences) and creates the training set.

package Cluster::Process_Intensities;
require Exporter;


use strict;
use warnings;

use FindBin qw($Bin);
use lib "$Bin";
use lib "$Bin/../";



use Readonly;
use Carp;
use List::MoreUtils qw (any);
use SGE_API;
use ExtraUtils qw(timestamp_file is_fasta);
use Submit_tools;
use File::Basename;
use Storable;
use Archive::Tar;
use Cwd qw(abs_path);
use List::Util qw (shuffle);
use File::Path; 



######


use Snapshot;
our @ISA    = ("Exporter");









#####################################################################################################################################

Readonly my $D_NAME_OF_UNIFIED_SAMPLE_INT           => 'unified_int_sample.txt';
Readonly my $D_SOFT_LINK_FOLDER                     => 'Soft_links/';
Readonly my $DOT                                    => q{.};
Readonly my $WILD_CARD_SYMBOL                       => '<X>';
Readonly my $D_SYMBOL                               => "$Snapshot::D_SYMBOL";
Readonly my $D_FOLDER_TRAIN                         => "$Snapshot::D_FOLDER_TRAIN";
Readonly my $D_LIFETIME                             => '10h';
Readonly my $LOCATION_OF_MAN_SCRIPTS                => "$Bin/Cluster/Remote_Scripts";
Readonly my $D_REFRESH_TIME                         => '0.5m';

Readonly my $D_REMOTE_SCRIPT_NAME                   => "$LOCATION_OF_MAN_SCRIPTS/"."pre_process_intensities.pl";

Readonly my $D_GZ_SUFFIX                             => ".gz";
Readonly my $D_GUNZIP                                => 'gunzip';
Readonly my $D_GZIP                                  => 'gzip';

##################################################################################################################################

#test();

sub test {
my @tiles = (35..38);
my $r_tiles = \@tiles;
    quick_snapshot_intensity ({
                                verbose => '1',
                                debug   => '1',
                                path_to_int_files => '/data/hannon/erlich/Alta-Cyclic/long_runs/BI.PE1.080723_SL-XAV_0004_FC304F8AAXX.4.run/',
                                output_dir =>'/data/hannon/erlich/Alta-Cyclic/long_runs/2008.09.16_19.32.02__Training.Folder_BI.PE1.080723_SL-XAV_0004_FC304F8AAXX.4.Alta_____Develop/',
                                lane => '4',
                                tiles => $r_tiles,
                                pos_to_seq => '/data/hannon/erlich/Alta-Cyclic/long_runs/'.
                                              '2008.09.16_19.32.02__Training.Folder_BI.PE1.080723_SL-XAV_0004_FC304F8AAXX.4.Alta_____Develop/Training_set_files/unified_seq_sample_for_blat.fa.PosToSeq.uniHash',

                             });
}



#######################################################################################
sub quick_snapshot_intensity {

    my ($arg_ref) = @_;

    my $verbose           = $arg_ref->{verbose};            #verbosity
    my $debug             = $arg_ref->{debug};
    my $path_to_int_files = $arg_ref->{path_to_int_files};
    my $output_dir        = $arg_ref->{output_dir};
    my $lane              = $arg_ref->{lane};
    my $r_tiles           = $arg_ref->{tiles};
    my $pos_to_seq        = $arg_ref->{pos_to_seq}; #file name!!
    my $zipped            = $arg_ref->{zipped};


    #zipped
    my $suffix_int_files = $D_SUFFIX_OF_INT_FILES;
    if ($zipped) {
        $suffix_int_files .= $D_GZ_SUFFIX;
    }

    $output_dir = "$output_dir/$D_FOLDER_TRAIN";




    my @int_files_fullname = 
        Snapshot::get_filename_in_illumina_format({path   => $path_to_int_files,
                                                   lane    => $lane,
                                                   tiles   =>$r_tiles,
                                                   suffix  => $suffix_int_files,
                                                   symbol  => $D_SYMBOL});



    my ($r_soft_links,$file_wildcard,$path_soft_links) 
                                                        = create_softlink_to_tiles (
                                                                        \@int_files_fullname,
                                                                        $zipped,
                                                         );



    my $uni_int_file = take_snapshot_int({
                                          verbose     => $verbose,
                                          output_dir  => $output_dir,
                                          pos_to_seq  => $pos_to_seq,
                                          int_files   => \@int_files_fullname,
                                          tiles       => $r_tiles,
                                          soft_links  => $r_soft_links,
                                          file_wildcard => $file_wildcard,
                                          debug        => $debug,
                                         });

    clean_soft_links($path_soft_links , $debug);


    return $uni_int_file;
}

####################################################################################################################################




 sub take_snapshot_int {
    my ($arg_ref) = shift;

    my $verbose                 =   $arg_ref->{'verbose'};
    my $pos_to_seq              =   $arg_ref->{'pos_to_seq'}; #this is hash file
    my $output_dir              =   $arg_ref->{'output_dir'} || '.';
    my $r_int_files_to_read     =   $arg_ref->{'int_files'}; #should be fullname`
    my @tiles                   =   @{$arg_ref->{'tiles'}};
    my $r_soft_links            =   $arg_ref->{'soft_links'};
    my $file_wildcard           =   $arg_ref->{'file_wildcard'};
    my $debug                   =   $arg_ref->{'debug'};




    #########################
    #the output file is defined as constant:
    my $outputfile = "$output_dir/$D_NAME_OF_UNIFIED_SAMPLE_INT";

    print "Taking sanpshot of intensity reads from the training library\n",
          "Reading training library\n"
        if ($verbose);


    print "\n",
          "--------------------------------------\n",
          "Names of intensity files:\n" if ($debug);

    foreach my $f (@$r_int_files_to_read) {
          print "$f\n" if ($debug);
    }

    print "--------------------------------------\n",
          "\n" if ($debug);

    ###########################


    my @starts =  $tiles[0];
    my @gaps   =  1;

    my $last_tile = $tiles[scalar @tiles-1];

    my @stops  =  $last_tile++;


    my $cluster_script = SGE_API->new();
    my ($counter,undef) = $cluster_script->set_range(\@starts,\@stops,\@gaps);

    print "Total number of permutations is $counter\n" if ($debug);


    $cluster_script->add_required_file({refer=> 'pre_process_intensities',
                                        name => $D_REMOTE_SCRIPT_NAME,
                                        place=> 'local'});

    $cluster_script->add_required_file({refer=> 'pos_to_seq',
                                        name => $pos_to_seq,
                                        place=> 'local'});

    $cluster_script->add_required_file({refer=>'int_file',
                                        name => "$file_wildcard",
                                        place=> 'local'
                                       });


    my $command_line = join (' ',
                                 '<pre_process_intensities>',
                                 '--intensity_file_to_read',      '<int_file>',
                                 '--pos_to_seq'            ,      '<pos_to_seq>',
                                 '--output_dir'            ,      $output_dir,
                                 '--tile_number'           ,      '<X>',
                            );


    $cluster_script->add_command($command_line);



    my $script_to_submit = $cluster_script->compile;
    $cluster_script->check_integrity;

    print "Generating script to send to the cluster: $script_to_submit\n" if ($debug);
    #if ($debug) {_print_script($script_to_submit)};


    my $qsub = Submit_tools->new();
    $qsub->life_time($D_LIFETIME);
    $qsub->range(1,scalar @tiles,1);
    $qsub->submit($script_to_submit);
    my $job_name = $qsub->get_job_id;
    my @stdout_files = $qsub->get_message_filenames;
    $qsub->automatic_pilot($D_REFRESH_TIME);
    $cluster_script->bye;


    my @int_files_to_unify = 
        read_messages_and_clean(\@stdout_files,$job_name,$verbose,$debug);

    my ($lines_in_the_unified_int_file, $fileout) = 
        unify(\@int_files_to_unify,$output_dir,$verbose);

    #There is a bug in the Illumina pipeline, where the position of the intensity
    # reads is not the same as the sequence reads (it seems due to rounding problems).
    #check_unified_correctly($lines_in_the_unified_int_file,$pos_to_seq);


    return $fileout;

}

sub check_unified_correctly {
    my ($lines_in_the_unified_int_file,$pos_to_seq) = @_;

    my $r_pos_to_seq = retrieve $pos_to_seq;

    my $lines_in_the_hash_pos_to_int =
        scalar keys %$r_pos_to_seq;

    confess "number of intensity reads for training is $lines_in_the_unified_int_file",
             " and is not equal to number of sequences for training: $lines_in_the_hash_pos_to_int","\n"
             if ($lines_in_the_unified_int_file != $lines_in_the_hash_pos_to_int);

    return 1;
}


sub unify {
    my ($r_int_files_to_unify,$output_dir,$verbose) = @_;


    my $fileout = $output_dir .'/'. $D_NAME_OF_UNIFIED_SAMPLE_INT;

    open my $FH_OUT, '>', $fileout or 
        confess "can't open $fileout: $!";


    my @FH_ARRAY;
    foreach my $int_file_to_unify (@$r_int_files_to_unify) {
        open my $FH, '<', $int_file_to_unify or 
            confess "can't open $int_file_to_unify: $!";

        push @FH_ARRAY, $FH;
    }

    @FH_ARRAY = shuffle @FH_ARRAY;
    my $line_counter = 0;

    while (scalar @FH_ARRAY > 0) {

        my $FH_to_read = shift @FH_ARRAY;
        my $line = <$FH_to_read>;

        if (not defined $line) {
            close $FH_to_read or 
                confess "can't close file: $!";
        }
        else {
            print {$FH_OUT} "$line";
            push @FH_ARRAY, $FH_to_read;
            $line_counter++;
        }

    }

    foreach my $int_file_to_unify (@$r_int_files_to_unify) {
        unlink $int_file_to_unify ||
           confess "can't open $int_file_to_unify: $!";
    }


    close $FH_OUT or 
        confess "can't close $fileout: $!";


    print "Unified ",scalar @$r_int_files_to_unify, " files. Total lines: $line_counter\n"
        if ($verbose);

    return ($line_counter,$fileout);



}

sub read_messages_and_clean {

    my ($r_files,$job_name,$verbose,$debug) = @_;

    my @filtered_intensity_files;

    my @message_files = map {abs_path $_ } @$r_files;

    my %lines_in_tile;

    if ($verbose) {
        print "Cleaning stdout files\n";
    }


    foreach my $file (@message_files) {
        open my $FH, '<', $file
            or confess "can't open $file: $!";

        my $tile;
        my $lines;

        while (<$FH>) {
            chomp;
            if (m/number of lines that remained to read:(\d+)/) {
                $lines = $1;
            }
            elsif (m/tile number is:(\d+)/) {
                $tile = $1;
            }
            elsif (not m/^\x2b/) {
                push @filtered_intensity_files, $_;
            }
        }

        $lines_in_tile{$tile} = $lines;

        close $FH;
        unlink $file ||
            croak "can't clean file: $file $!\n";
        if ($verbose) {print "$file cleaned\n"};
    }


    if ($debug) {
        print "Tile and remained lines:\n";
        foreach my $tile (sort {$a<=>$b} keys %lines_in_tile) {
            print "$tile\t",$lines_in_tile{$tile},"\n";
        }

    }


    if ($debug) {
        print "-------------------------------\n",
              "names of intensity files to read:\n ";
        print "$_\n" foreach @filtered_intensity_files;

        print "Total: ",scalar @filtered_intensity_files,"\n",
              "-------------------------------\n";
    }

    confess "The number of filtered_intensity_files is ",scalar @filtered_intensity_files,
            "The number of messages files is ", scalar @message_files,"\n",
            "They should be the same number!\n"
        if (scalar @filtered_intensity_files != scalar @message_files);


    return @filtered_intensity_files;

}

############################# take_snapshot_int #################################################################################


sub create_softlink_to_tiles {

    my ($r_int_files_to_read,$zipped) = @_;

    my @soft_links;

    if (not -d $D_SOFT_LINK_FOLDER) {
        mkdir $D_SOFT_LINK_FOLDER || confess "can't create $D_SOFT_LINK_FOLDER: $!";
    }

    my $path = abs_path "./$D_SOFT_LINK_FOLDER";



    my $new_file_template;
    my $changed_area;

    foreach  my $file (@$r_int_files_to_read) {

        my $new_file = basename $file;
        $file =~ m/\d+$D_SYMBOL(\d+)$D_SYMBOL/;


        my $number_with_zeros = "$1";
        my $number_without_zeros = $number_with_zeros;
        $number_without_zeros    = $number_without_zeros + 0;

        $changed_area = "$DOT$number_without_zeros";
        if ($zipped) {
            $changed_area .= $D_GZ_SUFFIX;
        }


        $new_file=~m/($D_SYMBOL$number_with_zeros$D_SYMBOL.*)/x;

        my $area_to_change = $1;
        confess "error. expected different pattern in intensity file names: $new_file\n"
            if (not defined $area_to_change);

        $new_file =~ s/$area_to_change/$changed_area/;

        $new_file_template = $new_file;

        $new_file = $path .'/' . $new_file;
        symlink $file,$new_file || confess "can't create soft line $file=>$new_file: $!\n";

        push @soft_links, $new_file;
    }


    $new_file_template =~ s/$changed_area/$DOT$WILD_CARD_SYMBOL/;

    if ($zipped) {
        $new_file_template .= $D_GZ_SUFFIX;
    }

    $new_file_template = $path .'/'. $new_file_template;


    return (\@soft_links,$new_file_template,$path);
# 
#     my $template_wildcard = "$lane$DOT$WILD_CARD_SYMBOL$DOT";
#     $new_file =~ s/$new_file_template/$template_wildcard/;

}

sub clean_soft_links {
    my ($path_to_soft_links, $debug) = @_;

    rmtree($path_to_soft_links, {verbose => $debug});
    return 1;
}




sub _print_script {

    my $filename = shift;
    print "\n",
          "-------------------------------------------\n",
    open my $FH,'<', $filename
        or confess "can't open $filename: $!\n";

    while (<$FH>) {
        print "$_";
    }
    print "-------------------------------------------\n",
          "\n";
    close $FH;
    return 1;
}



1;