#! /usr/bin/env perl

#Copyright 2008 Cold Spring Harbor Laboratory.
# Written by Yaniv Erlich (erlich@cshl.edu)


package karma;
use strict;
use warnings;
use Readonly;
use Pod::Usage;
use Carp;
use FindBin qw ($Bin);
use Archive::Tar;
use Getopt::Long;
use Cwd qw (abs_path getcwd);
use lib $Bin;
use lib "$Bin/Cluster";
use Snapshot;
use SGE_API;
use Submit_tools;
use File::Basename;
use Time::localtime;
use Deconv;
use Getopt::Long;
use Storable;




Readonly my $D_SUFFIX_OF_HASH             => "hash";
Readonly my $D_GZ_SUFFIX                  => ".gz";
Readonly my $SPACE                        => q{ };
Readonly my $DOT                          => q{.};
Readonly my $WILD_CARD_SYMBOL             => '<X>';
Readonly my $D_SUFFIX_OF_INT_FILES        => "$Snapshot::D_SUFFIX_OF_INT_FILES";
Readonly my $D_SYMBOL                     => "$Snapshot::D_SYMBOL";
Readonly my $D_CALLER_FILE_NAME           => "caller.pl";
Readonly my $D_CALLER_LIFETIME            => '48h';
Readonly my $D_REFRESH_TIME               => '2m';
Readonly my $D_NUMBER_OF_LANES            => '1-8';
Readonly my $D_SOFT_LINK_FOLDER           => 'Soft_links/';
Readonly my $D_SUFFIX_CALLER              => '_AltaCyclic_seq.txt';
Readonly my $D_SUFFIX_KARMA               => '_AltaCyclic_seq.all.txt';


##################################################################

# NOTICE: Deafault is for GAII
Readonly my $D_NUMBER_OF_TILES            => '1-100';


#####################################################################

#Readonly my $REMOTE_SCRIPT_NANE => 'caller.pl';
Readonly my @MANDATORY_SCRIPTS   => qw (Model3.pm
                                        Model.pm
                                        caller.pl
                                        svd);



my $location;
if (-e "$Bin/Cluster/Remote_Scripts/$D_CALLER_FILE_NAME") {
    $location = "$Bin/Cluster/Remote_Scripts/";
}
else {
    confess "can't find $D_CALLER_FILE_NAME\n";
}



Readonly my $LOCATION_OF_MAN_SCRIPTS         => $location;

# NOTICE: tiles should be with gap of 1!!!!
#^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ Headers ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

my $object = get_user_params();
scenario($object);













################################################# Scenario ################################
sub scenario {
    my $self = shift;
  #  $self->test_load_all_req_params;

    if ($self->flag_verbose_get) {
        print "All requested parameters were loaded\n";
    }
    if ($self->flag_debug_mode_set) {
        $self->debug_SVM_tar_package;
    }
    $self->genereate_list_of_files_to_read;
    $self->generate_command_line;

    if ($self->flag_debug_mode_set) {
        $self->debug_print_parameters;
    }
    if ($self->flag_verbose_set) {
        print $self->get_command_line,"\n";
    }

   foreach my $lane (@{$self->get_lanes_to_read_r}) {
        if ($self->flag_verbose_get) {
            print "Reading lane $lane\n";
        }
        $self->create_softlink_to_tiles($lane);
        $self->connect_to_SGE_API($lane);
        $self->submit;
        $self->unify_results($lane);
        $self->clear_cluster_script;
        $self->clear_cluster_object;
    }
    $self->stdout_clean;

    system ("touch calling.finished");
}




################################################# Get user paramertes ################################
sub get_user_params {

    my $self = karma->new();

    my @random_walk;
    my $path_to_intensity_files;
    my $lanes;
    my $tiles;
    my $flowcell_name;
    my $SVM_tar_location;
    my $number_of_cycles;
    my $verbose;
    my $debug;
    my $zipped;
    my $auto;
    my $bring_files_to = './';

    my $HELP   = 0;
    my $MANUAL = 0;



    GetOptions( "random_walk=f{2}"         => \@random_walk,
                "path_to_intensity_files=s" => \$path_to_intensity_files,
                "number_of_cycles=i"       => \$number_of_cycles,
                "verbose"                  => \$verbose,
                "SVM_tar_location=s"              => \$SVM_tar_location,
                "flowcell_name=s"          => \$flowcell_name,
                "lanes=s"                  => \$lanes,
                "tiles=s"                  => \$tiles,
                "debug"                    => \$debug,
                "zipped"                   => \$zipped,
                "auto"                     => \$auto,

                "help"                     =>\$HELP,
                "man"                      =>\$MANUAL,

               );

    pod2usage(1) if $HELP;
    pod2usage(-verbose => 2 ) if $MANUAL;


    if ($auto) {
        load_hash($self,$lanes,$tiles);
        return $self;
    }

    my ($p1,$p2) = @random_walk;


        if (
            (not defined $p1 or not defined $p2)        or
            (not defined $number_of_cycles)             or
            (not defined $path_to_intensity_files)      or
            (not defined $SVM_tar_location)             or
            (not defined $lanes)                        or
            (not defined $tiles)
           )                                                {
            print "Not all mandatory parameters were loaded\n";
            pod2usage(1);
        }

    #__________________________________________________________________
    if ($verbose) {
        $self->flag_verbose_set;
    } 

    if ($debug) {
        $self->flag_debug_mode_set;
    }
    my @tiles_a = _convert_range_to_list($tiles);
    my @lanes_a = _convert_range_to_list($lanes);

    $self->set_lanes_to_read_r(\@lanes_a);
    $self->set_tiles_to_read_r(\@tiles_a);
    $self->set_p1_p2($p1,$p2);
    $self->set_number_of_cycles($number_of_cycles);
    $self->set_SVM_files_in_tar_package($SVM_tar_location);
    $self->set_intensity_path($path_to_intensity_files);

    if (defined $flowcell_name) {
        $self->set_flowcellname($flowcell_name);
    }
    if (defined $zipped) {
        $self->set_intensity_files_are_zipped;
    }
    my $folder = $self->genereate_output_dir($bring_files_to);

    if ($self->flag_verbose_get) {
        print "Create a new folder in $folder\n";
    }

    return $self;
}

sub load_hash {
            
    #
    #loading hash file from brave_heart
    #


    my ($self,$lanes,$tiles) = @_;
    my @lanes_a = undef;
    my @tiles_a = undef;

    #check if the user defined specific tiles or lanes to read
    if (defined $lanes) {
        @lanes_a = _convert_range_to_list($lanes);
    }
    else {
        @lanes_a = _convert_range_to_list($D_NUMBER_OF_LANES);
    }

    if (defined $tiles) {
        @tiles_a = _convert_range_to_list($tiles);
    }
    else {
        @tiles_a = _convert_range_to_list($D_NUMBER_OF_TILES);
    }

    my @files_with_hash_suffix = glob "*.$D_SUFFIX_OF_HASH";
    confess "Identified more than one hash table: @files_with_hash_suffix"
        if (scalar @files_with_hash_suffix > 1);
    confess "Can't find any hash file from brave_heart"
        if (scalar @files_with_hash_suffix < 1);


    my $hash_file = shift @files_with_hash_suffix;
    my $r_hash_brave_heart = retrieve($hash_file);





    #setting the parameters
    if ($r_hash_brave_heart->{verbose}) {
        $self->flag_verbose_set;
    }

    if ($r_hash_brave_heart->{debug}) {
        $self->flag_debug_mode_set;
    }

    $self->set_lanes_to_read_r(\@lanes_a);
    $self->set_tiles_to_read_r(\@tiles_a);

    $self->set_p1_p2($r_hash_brave_heart->{best_p1},
                     $r_hash_brave_heart->{best_p2},
                    );
    $self->set_number_of_cycles($r_hash_brave_heart->{number_of_cycles});
    $self->set_SVM_files_in_tar_package($r_hash_brave_heart->{SVM_tar_location});
    $self->set_intensity_path($r_hash_brave_heart->{path_to_intensity_files});

    if ($r_hash_brave_heart->{flowcellname}) {
        $self->set_flowcellname($r_hash_brave_heart->{flowcellname});
    }
    if ($r_hash_brave_heart->{zipped}) {
        $self->set_intensity_files_are_zipped;
    }
    my $folder = $self->genereate_output_dir('./');
}

################################################# New object ################################
sub new {
        #Create a new object
	my $classname = shift;
	my $self = {};
	bless $self,$classname;
        $self->add_mandatory_files;
	return $self;
}


sub add_mandatory_files {
    my $self = shift;

    my $folder = $LOCATION_OF_MAN_SCRIPTS;

    foreach my $script_name (@MANDATORY_SCRIPTS) {

        my $filename = abs_path ("$folder/$script_name");
        confess "can't find mandatory file $filename"
            if (not -e $filename);

        my $key      = fileparse ($filename,qr/\.p./); #will process .pl and .pm
        $self->{mandatory_files}->{$key}->{name} = $filename;
    }

    return 1;
}
sub get_mandatory_files {
    my $self = shift;

    my @filenames;
    my @file_keys = keys (%{$self->{mandatory_files}});
    foreach my $key (@file_keys) {
        push @filenames, $self->{mandatory_files}->{$key}->{name};
    }
    return (\@file_keys,\@filenames);
}




sub genereate_output_dir {
    my $self = shift;
    my $output_dir = shift || 
        confess "genereate_output_dir error\n";

    $output_dir = $self->_auto_output_lib($output_dir);

    if (not -d $output_dir) {
        mkdir $output_dir ||
            confess "can't create directory $output_dir\n";
    }

    $output_dir = abs_path $output_dir;
    $self->{file}->{output_dir} = $output_dir;
    return $self->{file}->{output_dir};
}
sub get_output_dir {
    my $self = shift;
    return $self->{file}->{output_dir};
}

sub DESTROY {
    my $self = shift;
}

##################################### Flags ########################################

sub flag_verbose_set {
    my $self = shift;
    $self->{flags}->{verbose} = 1;
    return $self->{flags}->{verbose};
}
sub flag_verbose_get {
    my $self = shift;
    return $self->{flags}->{verbose};
}
sub flag_generated_list_of_files_set {
    my $self = shift;
    $self->{flags}->{generated_a_list_of_intensity_files} = 1;
    return $self->{flags}->{generated_a_list_of_intensity_files};
}

sub flag_generated_list_of_files_get {
    my $self = shift;
    return $self->{flags}->{generated_a_list_of_intensity_files};
}

sub flag_debug_mode_set {
    my $self = shift;
    $self->{flags}->{debug_mode} = 1;
    return $self->{flags}->{debug_mode};
}
sub flag_debug_mode_get {
    my $self = shift;
    return $self->{flags}->{debug_mode};
}


###################################### Debug #############################################
sub debug_SVM_tar_package {
    my $self = shift;
    my $tar_package = $self->get_SVM_files_in_tar_package || 
        confess "no Tar package is set\n";

    my $tar = Archive::Tar->new;
    $tar->read($tar_package);
    my @list_of_files = $tar->list_files;
    print "\n-----------------------------\n",
          "List of files in $tar_package:\n";
    print "$_\n" foreach @list_of_files, "\n",
          "Total files:", scalar @list_of_files,
          "\n-----------------------------\n";

    return 1;
}


sub debug_print_parameters {
    my $self = shift;
    my @p1_p2 = $self->get_p1_p2;
    my $random_walk = join $SPACE, @p1_p2;
    my $number_of_cycles = $self->get_number_of_cycles;
    my $SVM_tar = $self->get_SVM_files_in_tar_package;
    my $flowcellname = $self->get_flowcellname;
    my $output_dir = $self->get_output_dir;
    my $r_lanes = $self->get_lanes_to_read_r;
    my $r_tiles = $self->get_tiles_to_read_r;

    my $lanes = _convert_list_to_range($r_lanes);
    my $tiles = _convert_list_to_range($r_tiles);

    my $verbose;
    if ($self->flag_verbose_get) {
        $verbose = 'Yes';
    }
    else {
        $verbose = 'No';
    }


    print "\n----------------------------\n"                ,
          "Parameters:\n"                                   ,
          "Random_walk: $random_walk\n"                     ,
          "Number of Cycles: $number_of_cycles\n"           ,
          "SVM tar file: $SVM_tar\n"                        ,
          "Flowcell name: $flowcellname\n"                  ,
          "Output dir: $output_dir\n"                       ,
          "Lanes: $lanes\n"                                 ,
          "Tiles: $tiles\n"                                 ,
          "Verbose: $verbose\n"                             ,
         "\n----------------------------\n";

    return 1;
}




########################################### User setting ############################

#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Mandatory Setting !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
sub set_p1_p2 {
    my $self = shift;
    confess "set_p1_p2 error\n"
        if (scalar @_ < 2);
    my ($p1,$p2) = @_;
    $self->{params}->{p1} = $p1;
    $self->{params}->{p2} = $p2;

    return ($self->{params}->{p1},$self->{params}->{p2});
}
sub get_p1_p2 {
    my $self = shift;
    return 
    if (
        (not exists $self->{params}->{p1}) or 
        (not exists $self->{params}->{p2})
       );

    return ($self->{params}->{p1},$self->{params}->{p2});
}

sub set_number_of_cycles {
    my $self = shift;
    my $p = shift ||
        confess "set_number_of_cycles error\n";
    $self->{config}->{number_of_cycles} =  $p;
    return $self->{config}->{number_of_cycles};
}
sub get_number_of_cycles {
    my $self = shift;
    return $self->{config}->{number_of_cycles}
}

sub set_lanes_to_read_r {
    my $self = shift;
    my $r_lanes = shift || 
        confess "set_lanes_to_read_r error\n";
    confess "lanes should be a reference to an array\n"
    if (ref $r_lanes ne 'ARRAY');

    $self->{param}->{r_lanes} = $r_lanes;
    return $self->{param}->{r_lanes};
}
sub get_lanes_to_read_r {
    my $self = shift;
    return $self->{param}->{r_lanes};
}

sub set_tiles_to_read_r {
    my $self = shift;
    my $r_tiles = shift || 
        confess "set_tiles_to_read_r error\n";
    confess "tiles should be a reference to an array\n"
    if (ref $r_tiles ne 'ARRAY');

    $self->{param}->{r_tiles} = $r_tiles;
    return $self->{param}->{r_tiles};
}
sub get_tiles_to_read_r {
    my $self = shift;
    return $self->{param}->{r_tiles};
}


sub set_intensity_path {
    my $self = shift;
    my $int_path = shift || 
        confess "set_intensity_path error\n";
    $int_path = abs_path ($int_path);
    $self->{param}->{int_path} = $int_path;
    return $self->{param}->{int_path};
}
sub get_intensity_path {
    my $self = shift;
    return $self->{param}->{int_path};
}

sub set_SVM_files_in_tar_package {
    my $self = shift;
    my $SVM_files_fullname = shift ||
        confess "set_SVM_files_in_tar_package error\n";

    if (not -e $SVM_files_fullname) {
        confess "can't find $SVM_files_fullname\n";
    }
    $SVM_files_fullname = abs_path $SVM_files_fullname;
    $self->{files}->{SVM_tar} = $SVM_files_fullname;
    return $self->{files}->{SVM_tar};
}



# (((((((((((((((((((((((((((((((((((OPTIONAL SETTING)))))))))))))))))))))))))
sub get_SVM_files_in_tar_package {
    my $self = shift;
    return $self->{files}->{SVM_tar};
}

sub set_intensity_files_are_zipped {
    my $self = shift;
    if ($self->flag_generated_list_of_files_get) {
        confess "can't set_intensity_files_are_zipped ",
                "after generation of list of files to read\n";
    }
    $self->{params}->{intensity_files_zipped} = 1;
    return $self->{params}->{intensity_files_zipped};
}

sub get_intensity_files_are_zipped {
    my $self = shift;
    return $self->{params}->{intensity_files_zipped};
}

sub set_flowcellname {
    my $self = shift;
    my $flowcellname = shift ||
        confess "set_flowcellname error\n";
    $self->{param}->{flowcellname} = $flowcellname;
    return $self->{param}->{flowcellname};
}
sub get_flowcellname{
    my $self = shift;
    return $self->{param}->{flowcellname};
}




############################################## List of Intensity Files ##########################
sub genereate_list_of_files_to_read {
    my $self = shift;
    my $r_lanes = $self->get_lanes_to_read_r;
    my $r_tiles = $self->get_tiles_to_read_r;
    my $path    = $self->get_intensity_path;
    my $suffix  = $D_SUFFIX_OF_INT_FILES;

    if ($self->get_intensity_files_are_zipped) {
        $suffix .= $D_GZ_SUFFIX;
    }

    foreach my $lane (@$r_lanes) {

        my @intensity_files_in_a_specific_lane;
        foreach my $tile (@$r_tiles) {

           my @tiles = ($tile);
           my @files_to_read =
                Snapshot::get_filename_in_illumina_format ({path=>$path,
                                                            lane=>$lane,
                                                            tiles=>\@tiles,
                                                            suffix=>$suffix,
                                                            symbol=>$D_SYMBOL,});


           my $file_to_read = shift @files_to_read; #only one element in the array!
           if (not -e $file_to_read) {
                croak "can't find $file_to_read\n";
           }
           push @intensity_files_in_a_specific_lane , $file_to_read;
           $self->set_intensity_file_by_lane_and_tile($lane,$tile,$file_to_read);

        }   #tile
        @{$self->{files}->{intensity_files_to_read}->{$lane}} = @intensity_files_in_a_specific_lane;
    }   #lane

    $self->flag_generated_list_of_files_set;

}

sub set_intensity_file_by_lane_and_tile {
    my $self = shift;
    my $lane = shift ||
        confess "set_intensity_files_to_read_by_lane error - need lane\n";
    my $tile = shift ||
        confess "set_intensity_files_to_read_by_lane error - need tile\n";
    my $file = shift ||
        confess "set_intensity_files_to_read_by_lane error - need file\n";

    $self->{files}->{intensity_by_lane_and_tile}->{$lane}->{$tile} = $file;
    return $self->{files}->{intensity_by_lane_and_tile}->{$lane}->{$tile};
}

sub get_intensity_file_by_lane_and_tile {
    my $self = shift;
    my $lane = shift ||
        confess "get_intensity_files_to_read_by_lane error - need lane\n";
    my $tile = shift ||
        confess "get_intensity_files_to_read_by_lane error - need tile\n";

    return $self->{files}->{intensity_by_lane_and_tile}->{$lane}->{$tile};
}


sub get_intensity_files_to_read_by_lane {
    my $self = shift;
    my $lane = shift || 
        confess "get_intensity_files_to_read_by_lane error - need lane\n";
    return @{$self->{files}->{intensity_files_to_read}->{$lane}};
}

sub unify_results {
    my $self = shift;
    my $lane = shift || 
        confess "get_intensity_files_to_read_by_lane error - need lane\n";

    my $r_tiles = $self->get_tiles_to_read_r;
    my $outfile = _generate_name_for_unified_seq_file($lane,$r_tiles);

    my $output_dir = $self->get_output_dir;


    my $current_dir = getcwd;
    chdir $output_dir;
        #from that point everything is in the output_dir

    my @files_to_unify = glob "*$D_SUFFIX_CALLER";

    open my $OUT,'>', $outfile || confess "can't open $outfile: $!";
    foreach my $file (@files_to_unify) {
        open my $IN, '<', $file || confess "can't open $file: $!";
        while (<$IN>) {
            print {$OUT} "$_";
        }
        close $IN;
        unlink $file || confess "can't clean $file: $!";
    }
    close $OUT;

    chdir $current_dir;
        #back to normal dir

    return 1;
}




############################################## Command Line #######################

sub generate_command_line {

    my $self = shift;
    my @p1_p2 = $self->get_p1_p2;
    my $random_walk = join $SPACE, @p1_p2;
    my $number_of_cycles = $self->get_number_of_cycles;
    my $SVM_tar = $self->get_SVM_files_in_tar_package;
    my $flowcellname = $self->get_flowcellname;
    my $output_dir = $self->get_output_dir;

    my $caller_name_in_sge_script_format = $D_CALLER_FILE_NAME;
    $caller_name_in_sge_script_format =~ s/.pl//;
    $caller_name_in_sge_script_format = "<$caller_name_in_sge_script_format>";

    my $command_line = join ($SPACE,
                                $caller_name_in_sge_script_format,
                                '--random_walk'             ,   $random_walk,
                                '--prob_estimate'           ,
                                '--svm_files'               ,   $SVM_tar,
                                '--bring_files_to'          ,   $output_dir,
                                '--number_of_cycles'        ,   $number_of_cycles,
                                '--intensity_file_to_call'  ,   '<int_file>'
                              );

    $command_line .= $SPACE;

    if ($self->flag_verbose_set) {
        $command_line .= "--verbose" .$SPACE;
    }

    if ($self->get_flowcellname) {
        $command_line .= "--flowcell_name" .$SPACE. $flowcellname. $SPACE;
    }

    $command_line .= $SPACE;

    $self->{command_line} = $command_line;
    return $self->{command_line};

}

sub get_command_line {
    my $self = shift;
    return $self->{command_line};
}


################################################## Cluster Methods ##################

sub connect_to_SGE_API {

# NOTICE: tiles should be with gap of 1!!!!
    my $self = shift;
    my $lane = shift ||
        confess "must give lane to connect_to_SGE_API\n";

    #range
    my $r_tiles    = $self->get_tiles_to_read_r;
    my @tiles      = sort {$a <=> $b} @$r_tiles;
    my @first_tile = ($tiles[0]);   #ugly code
    my $size_tiles = scalar @tiles;
    my @last_tile  = ($tiles[$size_tiles-1]);

    _check_tiles_integrity($tiles[0],$tiles[$size_tiles-1],$r_tiles);   #end of ugly code
    #wild card


    #command_line
    my $command_line = $self->get_command_line;


    #range
    my $cluster_script = SGE_API->new;

    my ($counter,undef) = $cluster_script->set_range(\@first_tile,\@last_tile);
    $self->set_number_of_cluster_units($counter);


    #files:
    my ($r_referes, $r_fullnames) = $self->get_mandatory_files;

    foreach my $i (0..scalar @$r_fullnames-1) {
        my $filename = $r_fullnames->[$i];
        my $refer    = $r_referes->[$i];

        $cluster_script->add_required_file( {refer=> $refer,
                                             name => $filename,
                                             place=> 'local'});
    }

    $cluster_script->add_required_file({refer=> 'int_file', 
                                        name => $self->get_wild_card_to_tile_filename,
                                        place=> 'local'});

    $cluster_script->add_required_file({refer=>'SVM_package',
                                        name => $self->get_SVM_files_in_tar_package,
                                        place=> 'local'});

    #$cluster_script->add_command("$D_CALLER_FILE_NAME "

    $cluster_script->add_command("$command_line");

    my $file = $cluster_script->compile;
    $cluster_script->check_integrity;

    $self->set_cluster_script($file);
    $self->set_cluster_script_object($cluster_script);
}


sub set_wild_card_to_tile_filename {
    my $self = shift;
    my $wildcard = shift || 
        confess "set_wild_card_to_tile_filename error\n";
    $self->{wildcard} = $wildcard;
    return $self->{wildcard};
}

sub get_wild_card_to_tile_filename {
    my $self = shift;
    return $self->{wildcard};
}


sub set_number_of_cluster_units {
    my $self = shift;
    my $cluster_units = shift ||
        confess "set_number_of_cluster_units error\n";
    $self->{cluster_units} = $cluster_units;
    return $self->{cluster_units};
}

sub get_number_of_cluster_units {
    my $self = shift;
    return $self->{cluster_units};
}

sub set_cluster_script {
    my $self = shift;
    my $script = shift || 
        confess "set_cluster_script error\n";
    $self->{file}->{cluster_script} = $script;
    return $self->{file}->{cluster_script};
}

sub get_cluster_script {
    my $self = shift;
    return $self->{file}->{cluster_script};
}

sub clear_cluster_script {
    my $self = shift;
    $self->{file}->{cluster_script} = undef;
    return 1;
}

sub set_cluster_script_object {
    my $self = shift;
    my $object = shift ||
        confess "set_cluster_script_object error";
    $self->{objects}->{cluster_script} = $object;
    return $self->{objects}->{cluster_script};
}

sub get_cluster_object {
    my $self = shift;
    return $self->{objects}->{cluster_script};
}

sub clear_cluster_object {
    my $self = shift;
    $self->{objects}->{cluster_script} = undef;
    return 1;
}


sub create_softlink_to_tiles {

    my $self = shift;
    my $lane = shift || 
        confess "create_softlink_to_tiles error\n";


    if (not -d $D_SOFT_LINK_FOLDER) {
        mkdir $D_SOFT_LINK_FOLDER || confess "can't create $D_SOFT_LINK_FOLDER: $!";
    }

    my $path = abs_path "./$D_SOFT_LINK_FOLDER";

    my $suffix = $D_SUFFIX_OF_INT_FILES;
    if ($self->get_intensity_files_are_zipped) {
        $suffix .= $D_GZ_SUFFIX;
    }

    my $new_file = undef;
    my $new_file_template = undef;
    foreach  my $tile (@{$self->get_tiles_to_read_r}) {

        my @tmp_tile = (1);
        my $old_file = $self->get_intensity_file_by_lane_and_tile($lane,$tile) ||
            confess "can't find source file to soft link\n";
        my @new_files =  
             Snapshot::get_filename_in_illumina_format ({path=>$path,
                                                         lane=>$lane,
                                                         tiles=>\@tmp_tile,
                                                         suffix=>$suffix,
                                                         symbol=>$D_SYMBOL,});

        $new_file = shift @new_files; #new_files is array with one element


        $new_file =~ m/($lane$D_SYMBOL)(0*1$D_SYMBOL)/;
        my $template_to_change = "$1$2";
        $new_file_template     = "$lane$DOT$tile$DOT";

        $new_file =~ s/$template_to_change/$new_file_template/;
        symlink $old_file,$new_file || confess "can't create soft line $old_file=>$new_file: $!\n";

        $self->add_soft_link($new_file);
    }

    my $template_wildcard = "$lane$DOT$WILD_CARD_SYMBOL$DOT";
    $new_file =~ s/$new_file_template/$template_wildcard/;
    $self->set_wild_card_to_tile_filename($new_file);
    return 1;
}

sub add_soft_link {
    my $self = shift;
    my $soft_link = shift || confess "add_soft_link error";
    push @{$self->{files}->{soft_links}}, $soft_link;
    return @{$self->{files}->{soft_links}};
}

sub get_all_soft_links{ 
    my $self = shift;
    return @{$self->{files}->{soft_links}};
}


sub submit {
    my $self = shift;
    my $submit = Submit_tools->new;
    $submit->life_time($D_CALLER_LIFETIME);

    $submit->range(1,$self->get_number_of_cluster_units,1);
    $submit->submit($self->get_cluster_script);
    my $job_name = $submit->get_job_id;
    my @stdout_files = $submit->get_message_filenames;
    $self->set_message_filenames(@stdout_files);
    $submit->automatic_pilot($D_REFRESH_TIME);
    ($self->get_cluster_object)->bye;
    return 1;
}

sub set_message_filenames {
    my $self = shift;
    my @files = @_;
    confess "set_message_filenames error"
        if (scalar @files < 1);
    @{$self->{files}->{stdout_files}} = @files;
    return @{$self->{files}->{stdout_files}};
}

sub get_message_filenames {
    my $self = shift;
    return @{$self->{files}->{stdout_files}};
}

sub stdout_clean {

    my $self = shift;
    my @files = $self->get_message_filenames;
    @files = map {abs_path $_ } @files;

    if ($self->flag_verbose_get) {
        print "Cleaning stdout files\n";
    }

    my $job_name = 'Sequencing-stdoutfiles';

    #if debug mode keep the files:
    if ($self->flag_debug_mode_get) {
        my $garb_dir_fullname = Deconv::_make_grabage_directory($job_name);
        my $tar = Archive::Tar->new();
        $tar->add_files(@files);
        $tar->write("$garb_dir_fullname/$job_name.tar",1);
        print "files were tarred in $garb_dir_fullname under $job_name.tar";
    }

    foreach my $file (@files) {
        unlink $file ||
            confess "can't clean file: $file $!\n";
        if ($self->flag_verbose_get) {print "$file cleaned\n"};
    }

    foreach my $soft_link ($self->get_all_soft_links) {
        unlink $soft_link ||
            confess "can't clean soft_link: $soft_link $!\n";
        if ($self->flag_verbose_get) {print "The soft_link t$soft_link cleaned\n"};
    }
}



################################################## Tests ##################
sub test {
    my @lanes = qw(8);
    my @tiles = (1..100);
    my $Karma = Karma->new();
    $Karma->set_flowcellname('TEST');
    #$Karma->set_intensity_files_are_zipped;
    #print $Karma->genereate_output_dir;
    $Karma->set_p1_p2(1,1);
    $Karma->set_intensity_path('/home2/hannon/erlich/SGE-API/step2/Int_files_from_the_classic_run');
    $Karma->set_lanes_to_read_r(\@lanes);
    $Karma->set_tiles_to_read_r(\@tiles);
    $Karma->set_number_of_cycles(50);
    $Karma->set_SVM_files_in_tar_package('../tmp/big.tar');

    $Karma->genereate_list_of_files_to_read;
    $Karma->debug_SVM_tar_package;
    $Karma->generate_command_line;
    $Karma->debug_print_parameters;
    print $Karma->get_command_line,"\n";



    foreach my $lane (@lanes) {
        $Karma->create_softlink_to_tiles($lane);
        $Karma->connect_to_SGE_API($lane);
        $Karma->submit;
        $Karma->clear_cluster_script;
        $Karma->clear_cluster_object;
    }

    $Karma->stdout_clean;
}


sub test_load_all_req_params {
    my $self=shift;

    if (
         ($self->get_p1_p2)                             and
         ($self->get_number_of_cycles)                  and
         ($self->get_SVM_files_in_tar_package)
        )                                                      {

        return 1;
    }
    confess "Not all mandatory parameters have been set\n"; 
    return;
}



##################### Aux #####################################

sub _check_tiles_integrity {
    my ($first,$last,$r_range) = @_;
    if (scalar @$r_range != ($last-$first+1)) {
        confess "The gap between the tiles to call must be 1\n";
    }
    return 1;
}

sub _convert_list_to_range {
    my ($r_list) = @_;
    my $first = $r_list->[0];
    my $last  = scalar @$r_list;
    $last = $r_list->[$last-1];
    my $range = "$first"."-"."$last";
    return $range;

}

sub _auto_output_lib {
    my $self = shift;
    my $user_dir  = shift;

    $user_dir =~ s/\/$//; # get rid of the last / if present in the dir

    if (not -d $user_dir) {
        mkdir $user_dir ||
            confess "can't mkdir $user_dir: $!\n";
    }

    my $flowcellname = $self->get_flowcellname || '';
    my $tm = localtime;
    my $folderName = sprintf("%04d_%02d_%02d_%02d_%02d_%02d_AltaCyclicResults_%s",
                             ($tm->year)+1900      ,($tm->mon)+1,
                              $tm->mday             ,$tm->hour,
                              $tm->min              ,$tm->sec,
                              $flowcellname);

    $folderName = "$user_dir/$folderName";

    my $copy_folderName = $folderName;
    my $count = 1;
    while (-d $copy_folderName) {
        $count++;
        $copy_folderName = "$folderName.$count";
    }

    return $copy_folderName;
}

sub _convert_range_to_list {
    my $range = shift;
    my ($start,$stop) = split /-/ , $range;
    if (not defined $stop) {
        $stop = $start;
    }
    my @array = ($start..$stop);
    return @array;
}

sub _generate_name_for_unified_seq_file {

    my ($lane,$r_tiles) = @_;

    my @tiles = @$r_tiles;
    my $first_tile = shift @tiles;
    my $last_tile  = pop   @tiles;

    my $filename = "s_$lane"."_$first_tile-$last_tile"."_$D_SUFFIX_KARMA";
    return $filename;
}

#Just to aviod one time usage of these two variables (Which throws error\n";
eval {$Snapshot::D_SUFFIX_OF_INT_FILES=$Snapshot::D_SYMBOL};

1;


__END__

=head1 NAME


karma - The base caller part of Alta-Cyclic

=head1 SYNOPSIS

karma.pl --auto [--tiles] [--lanes] 
 
karma.pl [MANDATORY PARAMETERS]   [OPTIONAL PARAMETERS] 

=cut

=head1 OPTIONS


  --auto                            Gets the parameters from brave_heart. Must be run under Karma folder.
				    Notice that for GAI machines you should specify the tile range. 
				    See the second example.

  Mandatory Parameters:
  --random_walk                     Expect to get [best p1] and [best p2] that were found by brave_heart
  --number_of_cycles                Number of cycles in the run
  --path_to_intensity_files         Path to the folder that contains the intensity files for reading
  --lanes                           Lanes to read 
  --tiles                           Tiles to read
  --SVM_tar_location                Location of the SVM package that was generated by brave_heart



  Optional Parameters:
  --zipped                          If the intensity files are zipped
  --flowcell                        The flowcell name. Just for record purposes
  --verbose                         Set verbosity.
  --debug                           Give extra messages and indicators. Stores the .o files that return from the cluster

  Help Options:
   --help                           Show a list of parameters.
   --man                            Print a long manual.



=cut


=head1 EXAMPLES

karma.pl --random_walk 0.9925 0.9975 --lanes 1 --number_of_cycles 78 --path_to_intensity /home8/mccombie1/080411_FREDDYKRUEGER_0004_FC2052PAAXX/Data/C1-78_Firecrest1.9.2_22-04-2008_delabast/ --zipped --SVM_tar_location 
/home2/hannon/erlich/SGE-API/step3.5/80cycles-4lane-2/2008_04_25_18_25_06_AltaCyclic_Training_1/SVM.tar --verbose --debug --tiles 1-100 --flowcell_name 080411_FREDDYKRUEGER_0004_FC2052PAAXX

karma.pl --auto

karma.pl --auto --tiles 1-200

karma.pl --auto --tiles 1-100 --lanes 1-6
=cut


=head1 DESCRIPTION

karma.pl is part of Alta-Cyclic base caller. The program takes the optimized value that were found by brave_heart.pl, the random walk parameters and the SVMs, and calling
the bases of the intensity files that are in the path (according to the lane and tile ranges). The returned sequences are in Illumina format including quality scores that 
are calculated by the SVM. 


=cut

=head1 BUGS AND LIMITATIONS

If you feel you have found a bug in karma.pl, please submit a bug report at erlich@cshl.edu

=cut

=head1 AUTHOR

Written By Yaniv Erlich <erlich@cshl.edu>

=cut

=head1 LISCENSE AND COPYRIGHT

Copyright © 2008 Cold Spring Harbor Laboratory.
This is free software.  You may redistribute copies of it under the terms in LICENSE file.  There is NO WARRANTY, to the extent permitted by law.

=cut
    
